#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import healthcheck_python.config as config


def start():
	"""
	Create instances and start the healthcheck pipeline
	"""
	if config.STARTED:
		return

	from healthcheck_python.api import HealthCheckApi
	from healthcheck_python.manager import HealthCheckManager
	from healthcheck_python.updater import HealthCheckUpdater

	api = HealthCheckApi(config.HOST, config.PORT, config.status_queue, daemon=True)
	updater = HealthCheckUpdater(config.process_queue, config.status_queue, daemon=True)
	manager = HealthCheckManager(config.message_queue, config.process_queue, daemon=True)

	api.start()
	updater.start()
	manager.start()
	config.STARTED = True



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import functools
import logging
import time
from multiprocessing import Process, Queue, cpu_count
from queue import Empty

import bottle
from setproctitle import setproctitle

from healthcheck_python.release import __version__


class HealthCheckApi(Process):
	"""
	API responder class
	Creates a bottle instance and reports the health status
	"""

	def __init__(self, host, port, status_queue, daemon=False):
		super().__init__()

		self._host = host
		self._port = port
		self._status_queue = status_queue
		self.daemon = daemon

		self._app = bottle.Bottle()
		self._app.install(HealthCheckApi.logging)
		self._app.queue = Queue()
		self._app.nb_workers = cpu_count()

		self._app.route('/', method="GET", callback=HealthCheckApi._index)
		self._app.route('/health', method="GET", callback=self._health)

	def __del__(self):
		self.terminate()

	def run(self):
		setproctitle(self.__class__.__name__)
		bottle.run(self._app, host=self._host, port=self._port, quiet=True)

	@staticmethod
	def _index():
		return f"Hello there! I'm healthcheck-python v{__version__}"

	def _health(self):
		"""
		Health check path
		/health
		:return: overall status str(boolean).
		:return: If verbose mode enabled, return a dict with details about every service
		"""
		is_verbose = "v" in bottle.request.query.keys()

		while True:
			try:
				status = self._status_queue.get(block=True, timeout=1)
				if time.time() - status[0] <= 0.5:
					status = status[1]
					break
			except Empty:
				status = {'status': False, 'services': {}}
				break

		if is_verbose:
			return status

		return {'status': status['status']}

	@staticmethod
	def logging(fn):
		"""
		Wrapper for send logs to logging module
		"""

		@functools.wraps(fn)
		def log(*args, **kwargs):
			actual_response = fn(*args, **kwargs)

			logging.info(
				"%s %s %s %s",
				bottle.request.remote_addr,
				bottle.request.method,
				bottle.request.url,
				bottle.response.status
			)

			return actual_response

		return log



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import multiprocessing as mp

import pytest

from healthcheck_python.manager import HealthCheckManager
from healthcheck_python.service.periodic_service import PeriodicService


@pytest.fixture(scope='module')
def input_queue():
	return mp.Queue()


@pytest.fixture(scope='module')
def output_queue():
	return mp.Queue()


@pytest.fixture(scope='module')
def manager_object(input_queue, output_queue):
	return HealthCheckManager(input_queue, output_queue)


def test_success(output_queue, manager_object):
	manager_object._process_message(
		{'type': PeriodicService, 'name': 'test_service', 'start_time': 1, 'end_time': 2, 'timeout': 3}
	)
	message = output_queue.get(block=True, timeout=0.1)
	assert message[1]['test_service']['_last_end'] == 2



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import importlib


def class_for_name(class_name: str):
	"""
	Import service file
	:param class_name: CamelCase class name from healthcheck_python.service
	:return: <class>
	"""
	module_name = importlib.import_module("healthcheck_python")
	submodule_name = getattr(module_name, "service")
	class_ = getattr(submodule_name, class_name)
	return class_



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import multiprocessing as mp
import time

import pytest

from healthcheck_python.api import HealthCheckApi
from healthcheck_python.release import __version__


@pytest.fixture(scope='module')
def queue():
	return mp.Queue()


@pytest.fixture(scope='module')
def api_object(queue):
	return HealthCheckApi("0.0.0.0", 1234, queue)


def test_index(api_object):
	assert api_object._index() == f"Hello there! I'm healthcheck-python v{__version__}"


def test_failed_response(api_object):
	response = api_object._health()
	assert response == {'status': False}


def test_success(queue, api_object):
	queue.put((time.time() + 5, {'status': True, 'services': {}}))
	time.sleep(0.1)
	response = api_object._health()
	assert response == {'status': True}



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import sys


class CircularQueue:
	"""
	A circular queue implementation
	only enqueues data and allow peeking the first added element
	"""

	def __init__(self, k: int):
		self.k = k
		self.queue = [0.0] * k
		self.head = -1
		self.tail = 0
		self.first_full = False

	def enqueue(self, data):
		"""
		Add new data to queue
		:param data: new data to add
		"""
		self.head = (self.head + 1) % self.k
		if self.first_full:
			self.tail += 1
		if self.head == self.k - 1:
			self.first_full = True
		self.queue[self.head] = data

	def __len__(self) -> int:
		if self.first_full:
			return self.k
		return self.head + 1

	def peek_tail(self):
		"""
		Get the data added earliest
		:return: earliest added data
		"""
		return self.queue[self.tail]

	def peek_head(self):
		"""
		Get the data added earliest
		:return: earliest added data
		"""
		return self.queue[self.head]

	def mean_nonzero(self) -> float:
		"""
		Find mean of nonzero objects
		:return: mean float
		"""
		if self.first_full:
			return self.k / sum(self.queue)

		total_sum = sum(self.queue[self.tail:self.head + 1]) + sys.float_info.epsilon
		return (self.head - self.tail + 1) / total_sum



import pytest

from healthcheck_python.utils.circular_queue import CircularQueue


@pytest.fixture(scope='function')
def queue():
	return CircularQueue(4)


def test_circular_queue(queue):
	queue.enqueue(1)
	queue.enqueue(2)
	queue.enqueue(3)
	queue.enqueue(4)
	assert queue.peek_tail() == 1
	queue.enqueue(5)
	assert queue.peek_tail() == 2
	queue.enqueue(6)
	assert queue.peek_tail() == 3


def test_mean(queue):
	queue.enqueue(2)
	queue.enqueue(3)
	queue.enqueue(4)
	assert 0.33 == pytest.approx(queue.mean_nonzero(), 0.1)
	queue.enqueue(5)
	assert 0.28 == pytest.approx(queue.mean_nonzero(), 0.1)



import pytest

from healthcheck_python.service import PeriodicService


@pytest.fixture(scope="function")
def service():
	service = PeriodicService("service1")
	service.add_new_point({'start_time': 1, 'end_time': 2, 'timeout': 3})
	return service


def test_serialize(service):
	_dict = service.serialize()
	assert _dict['_last_start'] == 1
	assert _dict['_last_end'] == 2
	assert _dict['_timeout'] == 3

	assert _dict['_queue']['queue'][0] == 1


def test_deserialize(service):
	_dict = service.serialize()

	new_service = PeriodicService.parse_from_dict(_dict)
	assert new_service._last_start == 1
	assert new_service._last_end == 2
	assert new_service._timeout == 3
	assert new_service._queue.head == 0
	assert new_service._queue.peek_head() == 1



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class BaseService:
	"""
	Base service template
	All services has to implement this
	"""

	def __init__(self, name):
		self.name = name

	def json(self) -> dict:
		"""
		Returns all attributes as dict
		:return: dict, all object attributes
		"""
		raise NotImplementedError()

	def serialize(self) -> dict:
		"""
		Serialize the object to a dict. serialize object instances as well
		:return: dict, all object attributes
		"""
		raise NotImplementedError()

	@staticmethod
	def parse_from_dict(_dict: dict):
		"""
		Create an object instance from the _dict serialized by self.serialize()
		:return: object
		"""
		raise NotImplementedError()

	def add_new_point(self, point: dict) -> None:
		"""
		Add new function call
		:param point: dict, new function call service
		"""
		raise NotImplementedError()

	def is_healthy(self, current_time: float = None) -> bool:
		"""
		Check if last call is within timeout limits
		:param current_time: time.time() object, Optional, check the status with specific time
		:return: boolean, service status
		"""
		raise NotImplementedError()



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import multiprocessing as mp

import pytest

import healthcheck_python
from healthcheck_python import config


@pytest.fixture(scope='module')
def queue():
	return mp.Queue()


def test_periodic_wo_fps(queue):
	config.message_queue = queue

	@healthcheck_python.periodic(service="service1", timeout=1)
	def test_function():
		x = 2 + 3

	test_function()
	call_args = queue.get(block=True, timeout=0.1)
	assert call_args['name'] == "service1"
	assert call_args['start_time'] == 0
	assert call_args['timeout'] == 1
	assert call_args['name'] == "service1"


def test_periodic_fps(queue):
	config.message_queue = queue

	@healthcheck_python.periodic(service="service1", calc_fps=True, timeout=1)
	def test_function():
		x = 2 + 3

	test_function()
	call_args = queue.get(block=True, timeout=0.1)
	assert call_args['name'] == "service1"
	assert call_args['start_time'] != 0
	assert call_args['timeout'] == 1


def test_fps(queue):
	config.message_queue = queue

	@healthcheck_python.fps(service="service1")
	def test_function():
		x = 2 + 3

	test_function()
	call_args = queue.get(block=True, timeout=0.1)
	assert call_args['name'] == "service1"
	assert call_args['start_time'] != 0



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import time

import healthcheck_python.config as config
from healthcheck_python.service.periodic_service import PeriodicService
from healthcheck_python.utils.pipeline import start


def periodic(_func=None, *, service='unknown', calc_fps=False, timeout=5):
	"""
	Periodic check decorator
	Add this to your periodically called functions
	:param _func: Wrapped function
	:param service: Service name. This name will be reported with API call
	:param calc_fps: Calculate fps on each run
	:param timeout: The timeout in seconds needed between to consecutive _func() calls
	before marking the service down
	:return: original return values of _func()
	"""
	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			start_time = time.time() if calc_fps else 0
			ret_val = func(*args, **kwargs)
			end_time = time.time()

			config.message_queue.put(
				{
					'type': PeriodicService, 'name': service,
					'start_time': start_time, 'end_time': end_time, 'timeout': timeout
				}
			)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)


def fps(_func=None, *, service='unknown'):
	"""
	FPS Calculation decorator
	Add this to your periodically called functions to calculate the fps
	:param _func: Wrapped function
	:param service: Service name
	:return: original return values of _func()
	"""
	start()

	def wrapper(func):
		@functools.wraps(func)
		def wrapper_func(*args, **kwargs):
			start_time = time.time()
			ret_val = func(*args, **kwargs)
			end_time = time.time()

			config.message_queue.put(
				{
					'type': PeriodicService, 'name': service,
					'start_time': start_time, 'end_time': end_time
				}
			)

			return ret_val

		return wrapper_func

	if _func is None:
		return wrapper

	return wrapper(_func)



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import multiprocessing as mp
import queue
import time

from setproctitle import setproctitle


class HealthCheckManager(mp.Process):
	"""
	HealthCheckManager
	Gets decorator messages and keeps track of process calls
	"""

	def __init__(self, message_queue, process_queue, daemon=False):
		super().__init__()
		self.message_queue = message_queue
		self.process_queue = process_queue
		self.daemon = daemon

		self.continue_running = True
		self.processes = {}

	def run(self):
		setproctitle(self.__class__.__name__)
		while self.continue_running:
			try:
				message = self.message_queue.get(block=True, timeout=0.1)
				if message is None:
					break
			except queue.Empty:
				continue

			self._process_message(message)

	def __del__(self):
		self.continue_running = False

	def _process_message(self, message):
		"""
		Process a single decorator message and put it into update queue
		:param message: A decorator message includes service name,
		functions start and end time and the timeout
		"""
		process_name = message['name']
		service = self.processes.get(process_name)
		if service is None:
			process_type = message['type']
			service = process_type(process_name)

		service.add_new_point(message)
		self.processes[process_name] = service

		self.process_queue.put((
			time.time(),
			{key: value.serialize() for key, value in self.processes.items()}
		))



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time

from healthcheck_python.service.base_service import BaseService
from healthcheck_python.utils.circular_queue import CircularQueue


class PeriodicService(BaseService):
	"""
	Periodic Service
	This service has to be updated periodically, otherwise it is marked as failed
	"""

	def __init__(self, name, timeout=10):
		super().__init__(name)
		self._timeout = timeout

		self._last_start = None
		self._last_end = None
		self._timeout = 999

		self._status = False

		self._fps = 0
		self._queue = CircularQueue(50)

	def json(self):
		"""
		Returns all attributes as dict
		:return: dict, all object attributes
		"""
		return {
			'status': self._status,
			'last_start': self._last_start, 'last_end': self._last_end, 'timeout': self._timeout,
			'fps': self._fps
		}

	def serialize(self):
		return dict(self.__dict__, **{'_queue': self._queue.__dict__, 'class': self.__class__.__name__})

	@staticmethod
	def parse_from_dict(_dict):
		new_service = PeriodicService(_dict['name'])
		new_queue = CircularQueue(_dict['_queue']['k'])
		new_queue.__dict__ = _dict['_queue']
		_dict.pop('_queue')
		new_service.__dict__ = _dict
		new_service._queue = new_queue
		return new_service

	def add_new_point(self, point):
		"""
		Add new function call
		:param point: dict, new function call service
		"""
		if point is None:
			return

		self._last_end = point['end_time']
		timeout = point.get("timeout")
		if timeout is not None:
			self._timeout = point['timeout']
		if point['start_time'] != 0:
			self._last_start = point['start_time']
			self._queue.enqueue(self._last_end - self._last_start)

	def is_healthy(self, current_time=None):
		"""
		Check if last call is within timeout limits
		:param current_time: time.time() object, Optional, check the status with specific time
		:return: boolean, service status
		"""
		if self._last_end is None:
			return False
		if current_time is None:
			current_time = time.time()

		self._fps = self._queue.mean_nonzero()

		self._status = current_time - self._last_end <= self._timeout
		return self._status



from setuptools import setup, find_packages

from healthcheck_python.release import __version__, __author__


def read_file(file_name):
	"""Read file and return its contents."""
	with open(file_name, "r") as f:
		return f.read()


def read_requirements(file_name):
	"""Read requirements file as a list."""
	reqs = read_file(file_name).splitlines()
	if not reqs:
		raise RuntimeError(
			f"Unable to read requirements from the {file_name} file"
			"That indicates this copy of the source code is incomplete."
		)
	return reqs


setup(
	name="healthcheck-python",
	version=__version__,
	url="https://github.com/cagdasbas/healthcheck-python",
	python_requires='>=3.6',
	description="Health Check API for multiprocessing python apps",
	long_description=read_file("README.md"),
	long_description_content_type="text/markdown",
	author=__author__,
	author_email="cagdasbs@gmail.com",
	packages=find_packages("."),
	include_package_data=True,
	install_requires=read_requirements("requirements.txt"),
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"Natural Language :: English",
		"License :: OSI Approved :: Apache Software License",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
	],
	license_files=("LICENSE",),
)



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from multiprocessing import Queue

STARTED = False
HOST = os.getenv("PY_HEALTH_CHECK_HOST", "0.0.0.0")
PORT = os.getenv("PY_HEALTH_CHECK_PORT", "8080")

if isinstance(PORT, str) and PORT.isdecimal() and 1 < int(PORT) < 65535:
	PORT = int(PORT)
else:
	PORT = 8080

message_queue = Queue()
process_queue = Queue()
status_queue = Queue()



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import multiprocessing as mp
import time

import pytest

from healthcheck_python.service import PeriodicService
from healthcheck_python.updater import HealthCheckUpdater


@pytest.fixture(scope='module')
def input_queue():
	return mp.Queue()


@pytest.fixture(scope='module')
def output_queue():
	return mp.Queue()


@pytest.fixture(scope='module')
def updater_object(input_queue, output_queue):
	return HealthCheckUpdater(input_queue, output_queue)


def test_parse(updater_object):
	service1 = PeriodicService("service1")
	processes = updater_object.parse_message({'service1_key': service1.serialize()})
	assert "service1_key" in processes.keys()
	assert isinstance(processes['service1_key'], PeriodicService)
	assert processes['service1_key'].name == "service1"


def test_success(output_queue, updater_object):
	service1 = PeriodicService("service1")
	service1.add_new_point({'start_time': 1, 'end_time': 2, 'timeout': 3})
	updater_object._processes = {'test_service': service1}

	updater_object._check_health()
	message = output_queue.get(block=True, timeout=0.1)
	assert not message[1]['status']

	service2 = PeriodicService("service2")
	service2.add_new_point({'start_time': 1, 'end_time': time.time(), 'timeout': 10000})
	updater_object._processes = {'test_service': service2}

	updater_object._check_health()
	message = output_queue.get(block=True, timeout=0.1)
	assert message[1]['status']



#  Copyright (c) 2021.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import multiprocessing as mp
import queue
import time

from setproctitle import setproctitle

from healthcheck_python.utils.utils import class_for_name


class HealthCheckUpdater(mp.Process):
	"""
	Health Check Updater
	Regularly tries to fetch latest process structure and updates health status
	Updates the overall status every 0.5 seconds
	"""

	def __init__(self, process_queue, status_queue, daemon=False):
		super().__init__()
		self._process_queue = process_queue
		self._status_queue = status_queue
		self.daemon = daemon

		self.continue_running = True
		self._processes = {}
		self._classes = {}

	def run(self):
		setproctitle(self.__class__.__name__)

		while self.continue_running:
			while True:
				try:
					message = self._process_queue.get(block=True, timeout=0.1)
				except queue.Empty:
					break
				if message is None:
					break

				if time.time() - message[0] <= 0.5:
					self._processes = self.parse_message(message[1])
					break

			self._check_health()

	def __del__(self):
		self.continue_running = False

	def parse_message(self, message) -> dict:
		"""
		Parse incoming struct to create own copy of process service
		:param message: dict with pickled values. Each value has to be a instance of BaseService
		:return: dict with object values
		"""
		processes = {}
		for key, data in message.items():
			service_class = self._classes.get(data['class'])
			if service_class is None:
				service_class = class_for_name(data['class'])
				self._classes[data['class']] = service_class
			data.pop('class')
			new_service = service_class.parse_from_dict(data)
			processes[key] = new_service
		return processes

	def _check_health(self):
		"""
		check every services ending time to current time
		Every service should ended within defined timeout interval
		Free the status queue and put the latest status.
		"""
		call_time = time.time()
		status = True
		for _, service in self._processes.items():
			service_status = service.is_healthy(call_time)
			if not service_status:
				status = False

		self._status_queue.put((
			time.time(),
			{
				'status': status,
				'services': {key: service.json() for key, service in self._processes.items()}
			}
		))
